import yaml
import os
import sys
import json
from datetime import datetime
import tempfile
from pprint import pprint
from .utils import (
    check_call,
    check_output,
    parse_compose_project,
    fixed_relative_context,
    extra_env,
    get_machine_options,
    slugify,
)
from compose.service import Service
from compose.cli.main import (
    project_from_options,
    TopLevelCommand,
    log_printer_from_project,
    ConvergenceStrategy,
    BuildAction,
)
import docker
from compose.cli import signals
from compose.service import BuildError
from compose.cli.command import Environment

# from .Pipeline import Pipelines
from compose.container import Container
from compose.config.errors import ConfigurationError
from contextlib import contextmanager
from .Config import parse_config
from .traefik import get_traefik_rules
from io import StringIO
from contextlib import redirect_stdout
import dockerpty
from dockerpty.pty import PseudoTerminal, RunOperation, ExecOperation


class DockerClient(docker.DockerClient):
    def __init__(self, api):
        self.api = api


class EmptyPwnMachine:
    pass


class PwnMachine:
    _project = None
    _cmd = None
    _client = None

    def __init__(self, pm_path):
        self.config_dir = pm_path
        config_file = os.path.join(pm_path, "config.yml")
        self.config = parse_config(self.config_dir, config_file)
        self.name = "pm"
        self.compose_files = [
            service["compose-file"] for service in self.config["services"].values()
        ]

    @property
    def client(self):
        if self._client is None:
            self._client = DockerClient(self.project.client)
        return self._client

    @property
    def project(self):
        if self._project is None:
            try:
                self._project = self.get_project(self.compose_files)
                self.apply_traefik_rules()
            except ConfigurationError as e:
                exit(e.msg)
        return self._project

    @property
    def cmd(self):
        if self._cmd is None:
            p = self.project
            cmd_options = {
                "--tlsverify": True,
                "--tlscacert": p.client.verify,
                "--tlscert": p.client.cert[0],
                "--tlskey": p.client.cert[1],
                "--host": p.client.base_url.replace("https", "tcp"),
                "--file": self.compose_files,
                "--project-name": self.name,
            }
            self._cmd = TopLevelCommand(p, cmd_options)
        return self._cmd

    @property
    def machine_name(self):
        return self.config["docker-machine"]

    @property
    def container_names(self):
        names = []
        for service in self.config["services"].values():
            names += service["containers"]
        return names

    @property
    def service_names(self):
        return list(self.config["services"].keys())

    @property
    def volume_names(self):
        names = []
        for service in self.config["services"].values():
            names += service["volumes"]
        return names

    @property
    def volumes(self):
        return self.project.volumes.volumes

    def get_project(self, compose_files):
        """ Parse the compose files """

        domains_env = {"PM_DOMAINS": json.dumps(self.config["domains"])}

        with extra_env(
            self.config["environment"], domains_env
        ), fixed_relative_context():
            project = parse_compose_project(
                self.config_dir, self.machine_name, self.name, compose_files
            )
        return project

    def apply_traefik_rule(self, service_name, service_config, rule_name):
        ssl = rule_name == "https"
        if service_config[rule_name]:
            extra_labels = get_traefik_rules(
                service_name, service_config[rule_name], ssl=ssl
            )
            service = self.project.get_service(service_config[rule_name]["container"])
            if "labels" in service.options:
                service.options["labels"].update(extra_labels)
            else:
                service.options["labels"] = extra_labels

    def apply_traefik_rules(self):
        for service_name, service_config in self.config["services"].items():
            self.apply_traefik_rule(service_name, service_config, "http")
            self.apply_traefik_rule(service_name, service_config, "https")

    def ps(self, services):
        self.cmd.ps(
            {"--quiet": False, "--services": False, "--all": False, "SERVICE": services}
        )

    def env(self):
        check_call(["docker-machine", "env", "--", self.machine_name])

    def get_exec(self, service_name, command):
        service = self.project.get_service(service_name)
        container_name = service.get_container(1).name

        r = self.client.containers.get(container_name).exec_run(command, demux=True)
        stdout, _ = r.output
        return stdout

    def exec(self, service_name, command, privileged=False):
        options = {
            "--index": "1",
            "SERVICE": service_name,
            "COMMAND": command[0],
            "ARGS": list(command[1:]),
            "-T": False,
            "--detach": False,
            "--privileged": privileged,
            "--user": None,
            "--env": {},
            "--workdir": None,
        }
        # will call exit()
        self.cmd.exec_command(options)

    def ssh(self, args, get_output=False):
        cmd = ["docker-machine", "ssh", self.machine_name, *args]
        if get_output:
            return check_output(cmd)
        else:
            return check_call(cmd)

    def start(self, service_names):
        try:
            self.project.up(service_names=service_names, renew_anonymous_volumes=True)
        except BuildError as e:
            exit(f"Build failed for {e.service.name}")

    def stop(self, service_names, timeout=10):
        self.project.stop(service_names=service_names, timeout=timeout)

    def restart(self, service_names, timeout=10):
        self.stop(service_names, timeout)
        self.start(service_names)

    def build(self, service_names):
        try:
            self.project.build(service_names=service_names)
        except BuildError as e:
            print(e)
            exit(f"Build failed for {e.service.name}")

    def backup(self, volume_name):
        backup_path = "~/PM_Backup"
        volume_path = self.client.volumes.get(volume_name).attrs["Mountpoint"]

        backup_name = datetime.now().strftime(f"{volume_name}_%s.tgz")
        backup_fullpath = os.path.join(backup_path, backup_name)
        print(f"Backuping {volume_name}")
        self.ssh(["mkdir", "-p", backup_path])
        cmd = [
            "tar",
            "czf",
            backup_fullpath,
            volume_path,
        ]
        self.ssh(cmd, get_output=True)
        print(f"Backup available here: {backup_fullpath}")

    def logs(self, service_names, follow, tail):
        containers = self.project.containers(service_names=service_names, stopped=True)

        nocolor = False

        tail = tail if tail == "all" else int(tail)
        log_args = {"follow": follow, "tail": tail, "timestamps": False}
        log_printer_from_project(
            self.project,
            containers,
            nocolor,
            log_args,
            event_stream=self.project.events(service_names=service_names),
        ).run()

    def mount(self, volume_name, mountpoint):
        def do_mount(remote_path, local_path):
            success = check_call(
                [
                    "docker-machine",
                    "mount",
                    f"{self.machine_name}:{remote_path}",
                    local_path,
                ]
            )
            if not success:
                exit(1)
            print(
                f"Volume '{volume_name}' mounted on '{local_path}'.\n"
                "Exit this shell (ctrl+d) to unmount.\nStarting ...\r",
                end="",
            )
            check_call([os.environ["SHELL"]], cwd=local_path)
            check_call(["umount", local_path])

        volume = self.volumes[volume_name]
        remote_path = volume.inspect()["Mountpoint"]

        if mountpoint is not None:
            return do_mount(remote_path, mountpoint)

        prefix = f"{self.machine_name}-{volume_name}-"
        with tempfile.TemporaryDirectory(prefix=prefix) as mountpoint:
            return do_mount(remote_path, mountpoint)

    def get_container_names_for_services(self, service_names):
        container_names = []
        for service_name in service_names:
            container_names.extend(
                self.config["services"][service_name].get("containers", [])
            )
        return container_names



import os
import click
import shutil
import yaml
from .PwnMachine.utils import check_output, check_call
from click._bashcomplete import get_completion_script


PM_COMMENT = "# PWN MACHINE"
PM_COMMENT_END = "# PWN MACHINE END"


def get_pm_offset(content):
    try:
        start = content.index(PM_COMMENT)
        end = content.index(PM_COMMENT_END) + len(PM_COMMENT_END) + 1  # \n
    except ValueError:
        return None
    return [start, end]


def update_rc(file_path, lines):
    with open(file_path, "r") as f:
        filerc = f.read()

    if PM_COMMENT in filerc:
        click.echo(f"pm is already installed in {file_path}")
        if not click.confirm("Do you want to update with the new configuration ?"):
            return

        offset = get_pm_offset(filerc)
        if offset is None:
            click.echo("Unable to find closing comment", err=True)
            raise click.Abort()
        filerc = filerc[: offset[0]] + filerc[offset[1] :]

    if filerc[-1] != "\n":
        filerc += "\n"

    lines = [PM_COMMENT, *lines, PM_COMMENT_END]
    filerc = filerc + "\n".join(lines) + "\n"

    with open(file_path, "w") as f:
        f.write(filerc)


def prepare_bash(pm_path):
    click.echo("bash detected")
    bash_autocomplete_path = os.path.join(pm_path, ".bash-autocomplete.sh")
    with open(bash_autocomplete_path, "w") as f:
        f.write(get_completion_script("pm", "_PM_COMPLETE", "bash"))
    bashrc_path = os.path.expanduser("~/.bashrc")

    lines = [
        f'export PM_PATH="{pm_path}"',
        f". {bash_autocomplete_path}",
    ]

    if click.confirm("Do you want pm to update your bashrc ?"):
        update_rc(bashrc_path, lines)
    else:
        click.echo("Add this to your bashrc to enable pm")
        click.echo("\n".join(lines))


def prepare_zshrc(pm_path):
    click.echo("zsh detected")
    zsh_autocomplete_path = os.path.join(pm_path, ".zsh-autocomplete.sh")
    with open(zsh_autocomplete_path, "w") as f:
        f.write(get_completion_script("pm", "_PM_COMPLETE", "zsh"))

    zshrc_path = os.path.expanduser("~/.zshrc")
    lines = [
        f'export PM_PATH="{pm_path}"',
        f". {zsh_autocomplete_path}",
    ]
    if click.confirm("Do you want pm to update your zshrc ?"):
        update_rc(zshrc_path, lines)
    else:
        click.echo("Add this to your zshrc to enable pm")
        click.echo("\n".join(lines))


def update_unknowrc(pm_path):
    click.echo("Unable to autoconfigure your shell")
    click.echo(f"Please set the env variable PM_PATH={pm_path} in your init file.")


def update_rc_files(pm_path):
    shell = os.environ.get("SHELL")
    if shell.endswith("bash"):
        prepare_bash(pm_path)
    elif shell.endswith("zsh"):
        prepare_zshrc(pm_path)
    # elif shell.endswith("fish")
    #    update_fishrc(pm_path)
    else:
        update_unknowrc(pm_path)


# update_rc_files("/home/bitk/bi.tk/")
# exit()


def get_machine_ip(machine_name):
    """
    run docker-machine ip {machine-name} and return the ip
    abort if the machine is not found
    """
    ip = check_output(["docker-machine", "ip", machine_name])
    if ip is None:
        click.echo(f"No machine with name {machine_name} can be found.", err=True)
        raise click.Abort()
    return ip.strip()


def check_ns(domain, machine_ip):
    """
    Check that the NS for the domain point to the machine_ip
    return True if it match, False if it doesn't
    """
    import dns.resolver

    def lookup():
        for ns_entry in dns.resolver.query(domain, "NS"):
            for a_entry in dns.resolver.query(ns_entry.target, "A"):
                if a_entry.address == machine_ip:
                    return True
        return False

    if not lookup():
        click.echo(
            f"The ip of the docker machine ({machine_ip}) and the nameserver ip for '{domain}' doesn't match."
        )
        click.confirm("Are you sure you want to continue ?", abort=True)


def copy_skel(pm_path):
    """
    copy default services to the destination folder
    """
    click.echo(f"Copying skel to {pm_path}")
    try:
        current_folder = os.path.dirname(__file__)
        shutil.copytree(os.path.join(current_folder, "skel"), pm_path)
    except Exception as e:
        click.echo(str(e), err=True)
        raise click.Abort()


def write_default_config(pm_path, config):
    """
    Write the yaml configuration file
    """
    config_path = os.path.join(pm_path, "config.yml")
    click.echo(f"Writing configuration file to {config_path}")
    with open(config_path, "w") as f:
        yaml.dump(config, f, default_flow_style=False, sort_keys=False)


def ask_config_info():
    pm_path = os.path.expanduser(
        click.prompt(
            "Where do you want to install PwnMachine ?",
            default="~/pwn-machine",
            type=click.Path(exists=False, writable=True),
        )
    )

    machine_name = click.prompt(
        "Please enter the name of your docker-machine.", default="pm"
    )
    machine_ip = click.prompt("Please enter your server public ip")
    domain = click.prompt("Please enter your domain name")
    check_ns(domain, machine_ip)
    email = click.prompt("Please enter the email address to use with Let's encrypt")

    config = {
        "version": 1.1,
        "docker-machine": machine_name,
        "domains": {domain: machine_ip},
        "environment": {
            "PDNS_API_KEY": os.urandom(16).hex(),
            "EMAIL": email,
            "LETSENCRYPT_URL": "https://acme-v02.api.letsencrypt.org/directory",
        },
        "services": {
            "traefik": {},
            "dns": {},
            "web": {
                "https": {"domains": [f"*.{domain}"]},
                "http": {"domains": [f"unsafe.{domain}"]},
            },
        },
    }
    return pm_path, config


def install_pwn_machine():
    pm_path, config = ask_config_info()
    copy_skel(pm_path)
    write_default_config(pm_path, config)
    update_rc_files(pm_path)
    click.echo(
        "PWN Machine is now installed, spawn new shell or set PM_PATH to get started."
    )




import os
import textwrap
import docker
import json
from tqdm import tqdm
from .utils import parse_compose_project, get_machine_options, print_stream
from compose.cli.main import BuildAction


def get_services_for_id(client, id):
    filters = {"label": [f"pm.pipeline.id={id}", "pm.pipeline.service.name"]}
    running_services = client.containers.list(filters=filters)
    return {s.labels.get("pm.pipeline.service.name"): s for s in running_services}


class DockerClient(docker.DockerClient):
    def __init__(self, api):
        self.api = api


class Service:
    def __init__(self, name, arguments):
        self.name = name
        self.arguments = arguments

    @classmethod
    def from_config(cls, name, config):
        if config is None:
            config = {}
        arguments = config.get("arguments", [])
        return cls(name, arguments)

    def __repr__(self):
        args = " ".join(f"<{a}>" for a in self.arguments)
        return f"""{self.name} {args}"""


class Pipeline:
    def __init__(self, compose_content, project, name, arguments, services):
        self.name = name
        self.id = os.urandom(4).hex()
        self.project = project
        self.services = services
        self.client = DockerClient(self.project.client)
        self.compose_content = compose_content
        self.arguments = arguments

    @classmethod
    def from_config(cls, compose_content, project, name, config):
        if config is None:
            config = {}

        arguments = config.get("arguments", [])
        services = {
            name: Service.from_config(name, s)
            for name, s in config.get("services", {}).items()
        }
        return cls(compose_content, project, name, arguments, services)

    def start(self, args):
        self.check_required_args(args)
        self.assert_runner()
        self.build_services(do_build=BuildAction.none)

        container_name = f"pm-pipeline-{self.name}-{self.id}"
        env = {
            "compose": self.compose_content,
            "pipeline": json.dumps(self.serialize(args)),
        }
        volumes = {
            "/var/run/docker.sock": {"bind": "/var/run/docker.sock", "mode": "rw"}
        }
        labels = {"pm.pipeline.name": self.name, "pm.pipeline.id": self.id}
        self.client.containers.run(
            "pm-pipeline-runner",
            name=container_name,
            labels=labels,
            auto_remove=True,
            detach=True,
            volumes=volumes,
            environment=env,
        )
        print(f'Pipeline "{self.name}" started in container {container_name}')

    def build_runner(self):
        print("Building pm-pipeline-runner")
        stream = self.project.client.build(
            path="/home/bitk/bi.tk/pipeline",
            tag="pm-pipeline-runner",
            quiet=False,
            rm=True,
        )
        print_stream(stream)

    def assert_runner(self):
        try:
            self.client.images.get("pm-pipeline-runner")
        except docker.errors.ImageNotFound:
            self.build_runner()

    def build(self, do_build=BuildAction.force):
        self.build_runner()
        self.build_services()

    def build_services(self, do_build=BuildAction.force):
        to_build = [s for s in self.project.services if s.name in self.services]
        for service in to_build:
            print(f"Building '{service.name}'")
            service.ensure_image_exists(do_build=do_build)

    def check_required_args(self, supplied_args):
        for service in self.services.values():
            for arg in service.arguments:
                if arg not in supplied_args:
                    exit(f"Argument '{arg}' is required for '{service.name}'")

    def serialize(self, arguments):
        services = list(self.services.keys())
        return {
            "name": self.name,
            "id": self.id,
            "services": services,
            "environ": arguments,
        }

    def __repr__(self):
        services = " | ".join(f"{s}" for s in self.services.values())
        return f"""{self.name}: {services}"""


class Pipelines:
    def __init__(self, compose_content, project, pipelines):
        self.project = project
        self.client = DockerClient(self.project.client)
        self.compose_content = compose_content
        self.pipelines = pipelines

    @classmethod
    def from_config(cls, machine_name, config):
        if config is None:
            config = {}

        compose_file = config.get("compose")
        try:
            compose_content = open(compose_file, "r", encoding="utf8").read()
        except:
            exit("compose file in pipeline need to be readable")

        project = parse_compose_project(machine_name, "pm_pipeline", [compose_file])
        pipelines = {
            name: Pipeline.from_config(compose_content, project, name, p)
            for name, p in config.get("pipelines", {}).items()
        }
        return Pipelines(compose_content, project, pipelines)

    def ps(self):
        filters = {"label": "pm.pipeline.name"}
        containers = self.client.containers.list(filters=filters)
        for container in containers:
            name = container.labels["pm.pipeline.name"]
            id = container.labels["pm.pipeline.id"]
            print(f"{name} [{id}]:")
            services = get_services_for_id(self.client, id)
            for service in self.pipelines[name].services:
                running = "" if service not in services else " [Running]"
                print(f"  - {service}{running}")

    def ls(self):
        for pipeline in self.pipelines.values():
            print(pipeline)

    def build(self, pipeline_names):
        for name, pipeline in self.pipelines.items():
            if pipeline_names and name not in pipeline_names:
                continue
            pipeline.build()

    def start(self, pipeline_name, args):
        self.pipelines[pipeline_name].start(args)



import click
from tabulate import tabulate

@click.group(name="volume")
def cli():
    """Manage Volumes"""


@cli.command()
def ls():
    """list all volumes"""
    print(tabulate([[name] for name in pm.volume_names], headers=["Volume"]))


@cli.command()
@click.argument("volume_names", nargs=-1, type=click.Choice(pm.volume_names))
def backup(volume_names):
    """backup a volume locally"""
    if volume_names is None:
        volume_names = pm.volume_names
    for name in volume_names:
        pm.backup(name)


@cli.command()
def restore():
    """restore a volume backup"""
    exit("NOT IMPLEMENTED")


@cli.command()
@click.argument("volume_name", type=click.Choice(pm.volume_names))
@click.argument(
    "mountpoint", required=False, type=click.Path(exists=True, file_okay=False)
)
def mount(volume_name, mountpoint=None):
    """Mount a remote volume locally, and start a shell.
       The volume is unmounted when the shell is closed. 
       
       If no mountpoint is specified, a tempdir will be created in /tmp
    """
    pm.mount(volume_name, mountpoint)



import click
import json
from io import StringIO
from contextlib import redirect_stdout
from tabulate import tabulate


def send_request(method, path, data=None):
    out = StringIO()
    args = [
        "-X",
        method,
        f"http://127.0.0.1:8081/{path.lstrip('/')}",
        "-H",
        f"X-API-Key: {pm.config['environment']['PDNS_API_KEY']}",
    ]
    if data:
        args += ["-d", json.dumps(data)]

    out = pm.get_exec("powerdns", ["curl", *args])
    try:
        return json.loads(out or '"success!"')
    except Exception as e:
        exit(out.decode())


def get(path):
    return send_request("GET", path)


def patch(path, data):
    return send_request("PATCH", path, data)


@click.group(name="dns")
def cli():
    """Manage DNS"""


@cli.command()
@click.argument("zone")
@click.argument("name")
@click.argument("type")
@click.argument("record")
@click.option("--ttl", help="DNS entry ttl", default=60, type=int)
def add(
    zone, name, record, type, ttl,
):
    """add a DNS entry"""
    zone_path = f"/api/v1/servers/localhost/zones/{zone}"
    new_record = {
        "rrsets": [
            {
                "name": name,
                "type": type,
                "ttl": ttl,
                "changetype": "REPLACE",
                "records": [{"content": record, "disabled": False},],
            }
        ]
    }
    print(patch(zone_path, new_record))


@cli.command()
@click.argument("zone")
@click.argument("name")
@click.argument("ips", nargs=-1)
def rebind(zone, name, ips):
    """
    Add a domain that reply a random ip from a list
    """
    if len(ips) < 2:
        exit("You need at least 2 ips")
    zone_path = f"/api/v1/servers/localhost/zones/{zone}"
    lua_ip_array = ",".join(f"'{ip}'" for ip in ips)
    lua_code = f'({{{lua_ip_array}}})[math.random({len(ips)})]'
    new_record = {
        "rrsets": [
            {
                "name": name,
                "type": "LUA",
                "ttl": "0",
                "changetype": "REPLACE",
                "records": [{"content": f'A "{lua_code}"', "disabled": False},],
            }
        ]
    }
    print(patch(zone_path, new_record))


@cli.command()
@click.argument("zone")
@click.argument("name")
@click.argument("type")
def rm(zone, name, type):
    """remove a DNS entry"""
    zone_path = f"/api/v1/servers/localhost/zones/{zone}"
    new_record = {
        "rrsets": [{"name": name, "type": type, "changetype": "DELETE", "records": []}]
    }
    print(patch(zone_path, new_record))


@cli.command()
def ls():
    """list all dns entries"""
    zones = get("/api/v1/servers/localhost/zones")
    to_print = []
    for zone in zones:
        entries = get(f"/api/v1/servers/localhost/zones/{zone['id']}")
        records = []
        for entry in entries["rrsets"]:

            for record in entry["records"]:
                records.append(
                    [zone["id"], entry["name"], entry["type"], record["content"]]
                )

        # sort records by type then by name
        to_print.extend(sorted(records, key=lambda i: (i[2], i[1])))
        # Spacing
        to_print.append([])

    print(tabulate(to_print, headers=["Zone", "Name", "Type", "Record"]))



#!/usr/bin/env python3
import click
import os
import subprocess
import shlex
import yaml
import sys
import click
import json
import os
import shutil
from glob import glob
from textwrap import dedent
from pprint import pprint
from .PwnMachine.PwnMachine import PwnMachine
from .PwnMachine.utils import hl, check_output
from .install import install_pwn_machine
import dns.resolver

base_plugin_folder = os.path.join(os.path.dirname(__file__), "cli")


class PluginLoader(click.MultiCommand):
    def __init__(self, *plugin_folders):
        super().__init__()
        self.plugin_folders = plugin_folders

    def list_commands(self, ctx):
        commands = []
        for plugin_folder in self.plugin_folders:
            commands.extend(
                os.path.basename(filename)[:-3]
                for filename in glob(os.path.join(plugin_folder, "*.py"))
            )
        return sorted(commands)

    def find_first_match(self, name):
        for plugin_folder in self.plugin_folders:
            fn = os.path.join(plugin_folder, name + ".py")
            if os.path.isfile(fn):
                return fn
        return None

    def get_command(self, ctx, name):
        ns = {"pm": pm}
        fn = self.find_first_match(name)
        if not fn:
            exit(f"ERROR: Invalid command {name}")
        try:
            with open(fn) as f:
                code = compile(f.read(), fn, "exec")
                eval(code, ns, ns)
        except FileNotFoundError:
            exit(f"ERROR: Invalid command {name}")
        if "cli" not in ns:
            click.echo(f"Error while loading {fn}, cli not found", err=True)
            raise click.Abort()
        return ns["cli"]


@click.group()
def setup_only():
    pass


@click.group()
def main_cli():
    pass


@main_cli.command()
def version():
    """Show current version"""
    from . import __version__

    print(__version__)


@setup_only.command()
def setup():
    """Setup pwn machine"""
    try:
        install_pwn_machine()
    except click.Abort:
        click.echo("Aborting", err=True)
        exit(1)


@main_cli.command()
def setup():
    """Setup pwn machine"""
    try:
        install_pwn_machine()
    except click.Abort:
        click.echo("Aborting", err=True)
        exit(1)


@main_cli.command()
def env():
    """Display the commands to set up the environment for the Docker client"""
    pm.env()


@main_cli.command()
@click.argument("command", nargs=-1)
def ssh(command):
    """SSH to the host"""
    pm.ssh([] if command is None else command)


if "PM_PATH" in os.environ and os.environ["PM_PATH"]:
    pm = PwnMachine(os.environ["PM_PATH"])

    @main_cli.command()
    @click.argument("services", nargs=-1, type=click.Choice(pm.service_names))
    @click.pass_context
    def ps(context, services):
        """List services"""
        pm.ps(services)

    plugin_folder = os.path.join(os.environ["PM_PATH"], "cli")
    cli = click.CommandCollection(
        sources=[main_cli, PluginLoader(base_plugin_folder, plugin_folder)]
    )
else:
    cli = setup_only


if __name__ == "__main__":
    if "PM_PATH" in os.environ and os.environ["PM_PATH"]:
        cli()
    else:
        click.echo("PM_PATH is not present is the environment.")
        if click.confirm(
            "Do you want to install pwn-machine ?", abort=True, default=True
        ):
            setup()



import click
import json
from io import StringIO
from contextlib import redirect_stdout
from tabulate import tabulate


@click.group(name="service")
def cli():
    """Manage Services"""

@cli.command()
def ls():
    """list all services"""
    print(tabulate([[name] for name in pm.service_names], headers=["Service"]))

@cli.command()
@click.argument("service_names", nargs=-1, type=click.Choice(pm.service_names))
def build(service_names):
    """Build services"""
    if not service_names:
        service_names = pm.service_names
    pm.build(pm.get_container_names_for_services(service_names))


@cli.command()
@click.argument("service_names", nargs=-1, type=click.Choice(pm.service_names))
def start(service_names):
    """Start services"""
    pm.start(pm.get_container_names_for_services(service_names))


@cli.command()
@click.option("-t", help="Timeout before killing the service.", default=10)
@click.argument("service_names", nargs=-1, type=click.Choice(pm.service_names))
def stop(service_names, t):
    """Stop services"""
    pm.stop(pm.get_container_names_for_services(service_names), t)


@cli.command()
@click.option("-t", help="Timeout before killing the service.", default=10)
@click.argument("service_names", nargs=-1, type=click.Choice(pm.service_names))
def restart(service_names, t):
    """Restart services"""

    pm.restart(pm.get_container_names_for_services(service_names), t)


@cli.command()
@click.option("-f", help="Follow log output.", is_flag=True)
@click.option(
    "--tail",
    nargs=1,
    default="all",
    help='Number of lines to show from the end of the logs (default "all")',
)
@click.argument("service_names", nargs=-1, type=click.Choice(pm.service_names))
def logs(service_names, f, tail):
    """Fetch the logs of a service"""
    pm.logs(pm.get_container_names_for_services(service_names), f, tail)



import click
import json
from io import StringIO
from contextlib import redirect_stdout
from tabulate import tabulate


@click.group(name="container")
def cli():
    """Manage containers"""


@cli.command()
def ls():
    """list all container"""
    to_print = []
    for service_name, service in pm.config["services"].items():
        first = True
        for container in service["containers"]:
            to_print.append([service_name if first else "", container])
            first = False
    print(tabulate(to_print, headers=["Service", "Container"]))

@cli.command()
@click.option(
    "--privileged", help="Give extended privileges to the command", is_flag=True
)
@click.argument("container_names", type=click.Choice(pm.container_names))
@click.argument("command", nargs=-1)
def exec(container_names, command, privileged):
    """Run a command in a running container"""
    if not command:
        command = ["bash"]
    pm.exec(container_names, command, privileged)


@cli.command()
@click.option("-f", help="Follow log output.", is_flag=True)
@click.option(
    "--tail",
    nargs=1,
    default="all",
    help='Number of lines to show from the end of the logs (default "all")',
)
@click.argument("container_names", nargs=-1, type=click.Choice(pm.container_names))
def logs(container_names, f, tail):
    """Fetch the logs of a container"""
    pm.logs(container_names, f, tail)


@cli.command()
@click.argument("container_names", nargs=-1, type=click.Choice(pm.container_names))
def build(container_names):
    """Build containers"""
    pm.build(container_names)


@cli.command()
@click.argument("container_names", nargs=-1, type=click.Choice(pm.container_names))
def start(container_names):
    """Start containers"""
    pm.start(container_names)


@cli.command()
@click.option("-t", help="Timeout before killing the service.", default=10)
@click.argument("container_names", nargs=-1, type=click.Choice(pm.container_names))
def stop(container_names, t):
    """Stop containers"""
    pm.stop(container_names, t)


@cli.command()
@click.option("-t", help="Timeout before killing the service.", default=10)
@click.argument("container_names", nargs=-1, type=click.Choice(pm.container_names))
def restart(container_names, t):
    """Restart containers"""

    pm.restart(container_names, t)



#!/usr/bin/env python
import os
from setuptools import setup, find_packages

path = os.path.dirname(__file__)
long_description = open(os.path.join(path, "README.md"), "r", encoding="utf8").read()

setup(
    name="pwn-machine",
    version="1.1",
    packages=find_packages(),
    description="Simple self hosting solution based on docker for bug hunters.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/yeswehack/pwn-machine",
    author="@BitK_",
    author_email="bitk@yeswehack.com",
    install_requires=[
        "click==7.1.2",
        "dnspython==1.16.0",
        "docker==4.2.2",
        "docker-compose==1.26.2",
        "dockerpty==0.4.1",
        "docopt==0.6.2",
        "paramiko==2.7.1",
        "pyaml==20.4.0",
        "requests==2.24.0",
        "tqdm==4.47.0",
        "tabulate==0.8.7",
    ],
    package_data={"skel": ["*"]},
    include_package_data=True,
    entry_points={"console_scripts": ["pm=pwnmachine.cli:cli"]},
)



import subprocess
import shlex
import io
import os
import json
from tqdm import tqdm
from pprint import pprint
from contextlib import contextmanager, redirect_stdout, redirect_stderr
from compose.service import Service
from compose.cli.main import project_from_options, Environment
from collections import ChainMap
import click

def hl(color, s):
    return click.style(s, fg=color)

def check_call(*args, **kwargs):
    try:
        subprocess.run(*args, check=True, **kwargs)
    except subprocess.CalledProcessError:
        return False
    except KeyboardInterrupt:
        return False
    return True


def check_output(*args, **kwargs):
    try:
        return subprocess.check_output(*args, **kwargs).decode("utf8")
    except subprocess.CalledProcessError:
        return None


def slugify(name):
    return name.replace("*.", "wildcard_").replace(".", "-")


def print_stream(stream):
    progress = {}
    for line in stream:
        line = json.loads(line)
        if "stream" in line:
            tqdm.write(line["stream"], end="")
        elif "status" in line:
            if "id" not in line:
                continue
            if line["id"] not in progress:
                progress[line["id"]] = tqdm(leave=False)
            p = progress[line["id"]]
            if "progress" not in line:
                p.display(line["status"])
            else:
                p.set_description(line["status"])
                p.display(line["status"] + ": " + line["progress"])
    for p in progress.values():
        p.close()
    print()


def get_machine_options(machine_name):
    output = check_output(["docker-machine", "config", "--", machine_name])
    if output is None:
        exit(f'Unknow machine "{machine_name}"')
    options = {}
    for quoted_line in output.strip().split("\n"):
        line = shlex.split(quoted_line)[0]
        if "=" in line:
            name, _, value = line.partition("=")
            if name == "-H":
                name = "--host"
            options[name] = value
        else:
            options[line] = True
    return options


def parse_compose_project(project_dir, machine_name, name, compose_files):
    """ Parse the compose files """
    machine_options = get_machine_options(machine_name)
    compose_options = {
        **machine_options,
        "--file": compose_files,
        "--project-name": name,
        "--project-directory": project_dir,
    }

    # out = io.StringIO()
    # with redirect_stderr(out):
    project = project_from_options(project_dir, compose_options)
    return project


@contextmanager
def monkey_patch(module, method_name, replacement):
    original = getattr(module, method_name)

    def handler(*args, **kwargs):
        return replacement(original, *args, **kwargs)

    try:
        setattr(module, method_name, handler)
        yield
    finally:
        setattr(module, method_name, original)


def extra_env(*extra_env):
    """ Add pwn-machine env to the docker-compose project"""

    def new_from_env_file(original, *args, **kwargs):
        env = original(*args, **kwargs)
        for extra in extra_env:
            env.update(extra)
        return env

    return monkey_patch(Environment, "from_env_file", new_from_env_file)


def fixed_relative_context():
    """Fix relative context in service docker-compose file """
    from compose.config import config

    def new_process_config_file(original, *args, **kwargs):
        config = original(*args, **kwargs)
        config_dir = os.path.dirname(config.filename)
        for service in config.config["services"].values():
            if "build" not in service:
                continue
            if isinstance(service["build"], str):
                fixed_path = os.path.join(config_dir, service["build"])
                service["build"] = os.path.realpath(fixed_path)
            elif "context" in service["build"]:
                fixed_path = os.path.join(config_dir, service["build"]["context"])
                service["build"]["context"] = os.path.realpath(fixed_path)
        return config

    return monkey_patch(config, "process_config_file", new_process_config_file)



from collections.abc import Iterable
from pprint import pprint
from contextlib import contextmanager
import yaml
import re
import os
import bcrypt
import ipaddress
from dataclasses import dataclass
from .utils import hl
from collections import namedtuple

_re_container_name = r"[a-zA-Z0-9][a-zA-Z0-9_.-]*"
re_container_name = re.compile(f"^({_re_container_name})$")
re_container_name_port = re.compile(
    f"^(?P<container>{_re_container_name}):(?P<port>[0-9]+)$"
)
re_auth = re.compile(r"^(?P<login>[^:]+):(?P<hash>.+)$")
VERSION = 1.1


def yaml_load(filename):
    try:
        with open(filename) as f:
            return yaml.safe_load(f)
    except yaml.parser.ParserError as e:
        exit(e)


@contextmanager
def keep_error_trail(name):
    try:
        yield
    except ParserError as e:
        e.trail = [name, e.trail]
        raise e


def as_list(obj):
    if isinstance(obj, list):
        return obj
    if obj is None:
        return list()
    return (obj,)


def get_ip(ip):
    assert_type(ip, str)
    try:
        if "/" in ip:
            return ipaddress.ip_network(ip, strict=False)
        else:
            return ipaddress.ip_address(ip)
    except ValueError as e:
        raise ParserError(str(e))


def get_domains(domain_list):
    for domain in domain_list:
        yield domain


class ParserError(Exception):
    def __init__(self, msg, trail=None):
        super().__init__(msg)
        self.msg = msg
        self.trail = [trail, None] if trail else None

    def __str__(self):
        if self.trail:
            return f"{hl('red', '[ERROR]')} {self.msg} in {self.get_trail()}"
        else:
            return f"{hl('red', '[ERROR]')} {self.msg}"

    def get_trail(self):
        def inner(trail):
            current, next_item = trail
            if next_item:
                return f"{hl('blue', current)}.{inner(next_item)}"
            return hl("blue", current)

        return inner(self.trail)


def assert_type(obj, valid_type):
    if not isinstance(valid_type, tuple):
        valid_type = (valid_type,)
    if not isinstance(obj, valid_type):
        valid = "' or '".join(hl("yellow", t.__name__) for t in valid_type)
        raise ParserError(
            f"Invalid type '{hl('yellow', type(obj).__name__)}', expected '{valid}'"
        )
    return obj


def get_regex(regex, str):
    try:
        match = next(regex.finditer(str))
        return match.groupdict()
    except StopIteration:
        raise ParserError(
            f"Invalid value '{hl('yellow', str)}' expected '{hl('yellow', regex.pattern)}'"
        )


def get_typed_value(config, key, expected_type, required=True):
    if key not in config:
        if required:
            raise ParserError(f"Missing required key '{hl('blue', key)}'")
        return None

    value = config.get(key)

    with keep_error_trail(key):
        assert_type(value, expected_type)
    return value


def parse_basic_auth(domain):
    basic_auth = as_list(
        get_typed_value(domain, "basic-auth", (list, str), required=False)
    )

    for entry in basic_auth:
        match = get_regex(re_auth, entry)
        if match["login"].startswith("~"):
            hash = bcrypt.hashpw(match["hash"].encode(), bcrypt.gensalt())
            yield f"{match['login'][1:]}:{hash.decode()}"
        else:
            yield f"{match['login']}:{match['hash']}"


def parse_ip_list(domain, type_name):
    allow_list = as_list(
        get_typed_value(domain, type_name, (list, str), required=False) or list()
    )

    with keep_error_trail(type):
        return [get_ip(ip) for ip in allow_list]


def parse_domain(domain_name, domain):
    with keep_error_trail(domain_name):
        result = {}
        if domain is None:
            domain = {}

        with keep_error_trail("basic-auth"):
            result["basic-auth"] = list(parse_basic_auth(domain))

        with keep_error_trail("ip-allow-list"):
            result["ip-allow-list"] = parse_ip_list(domain, "ip-allow-list")

        with keep_error_trail("ip-deny-list"):
            result["ip-deny-list"] = parse_ip_list(domain, "ip-deny-list")

        return result


def find_exposed(compose):
    found = False
    for service_name, service in compose.get("services", {}).items():
        labels = service.get("labels")
        if labels is None:
            continue
        if isinstance(labels, list):
            labels = {l.partition("=")[0]: l.partition("=")[2] for l in labels}
        for name, value in labels.items():
            if name == "pm.http.port":
                if found:
                    raise ParserError(
                        f"Multiple {hl('yellow', 'pm.http.port')} labels found"
                    )
                found = {"container": service_name, "port": value}

    if not found:
        raise ParserError(
            f"Unable to find a container with label {hl('yellow', 'pm.http.port')}",
            "labels",
        )

    return found


def parse_web(http, compose_file):
    result = {}

    compose = yaml_load(compose_file)

    shorter_name = "/".join(compose_file.split("/")[-3:])
    with keep_error_trail(f"[...{shorter_name}]"):
        container_port = find_exposed(compose)

    result.update(container_port)

    domain_list = get_typed_value(http, "domains", (dict, list))
    if isinstance(domain_list, list):
        domain_list = {name: {} for name in domain_list}
    result["domains"] = {
        domain_name: parse_domain(domain_name, domain)
        for domain_name, domain in domain_list.items()
    }

    return result


def get_info_from_compose(compose_file):
    config = yaml_load(compose_file)

    services = list(config.get("services", {}).keys())
    volumes = []
    for volume in config.get("volumes", {}).values():
        if volume and "name" in volume:
            volumes.append(volume["name"])
    return services, volumes


def parse_service(config_dir, name, service):
    if service is None:
        service = {}
    with keep_error_trail(name):
        result = {}

        if "compose-file" in service:
            compose_path = get_typed_value(service, "compose-file", str)
        else:
            compose_path = f"./services/{name}/docker-compose.yml"

        result["compose-file"] = os.path.realpath(
            os.path.join(config_dir, compose_path)
        )
        containers, volumes = get_info_from_compose(result["compose-file"])
        result["containers"] = containers
        result["volumes"] = volumes
        with keep_error_trail("http"):
            http = get_typed_value(service, "http", dict, required=False)
            result["http"] = parse_web(http, result["compose-file"]) if http else None

        with keep_error_trail("https"):
            https = get_typed_value(service, "https", dict, required=False)
            result["https"] = (
                parse_web(https, result["compose-file"]) if https else None
            )
        return result


def parse_domains(domains):
    result = {}
    for domain, ip in domains.items():
        assert_type(domain, str)
        with keep_error_trail(domain):
            assert_type(ip, str)
        result[domain] = ip
    return result


def parse(config_dir, config):
    result = {}
    version = get_typed_value(config, "version", float)
    if version > VERSION:
        raise ParserError(
            f"Configuration file is from newer version {version} > {VERSION}"
        )

    domains = get_typed_value(config, "domains", dict)
    with keep_error_trail("domains"):
        if not domains:
            raise ParserError(f"You need at least one domain")
        result["domains"] = parse_domains(domains)
    result["docker-machine"] = get_typed_value(config, "docker-machine", str)
    services = get_typed_value(config, "services", dict)

    with keep_error_trail("services"):
        result["services"] = {
            name: parse_service(config_dir, name, service)
            for name, service in services.items()
        }
    result["environment"] = (
        get_typed_value(config, "environment", dict, required=False) or {}
    )
    return result


def parse_config(config_dir, filename):
    yaml_config = yaml_load(filename)
    try:
        config = parse(config_dir, yaml_config)
        return config
    except ParserError as e:
        exit(str(e))



from .utils import slugify



def get_traefik_rules(service_name, http_config, ssl=False):
    labels = {}
    port = http_config["port"]
    labels["traefik.enable"] = "true"
    labels["traefik.docker.network"] = "traefik"

    r = "traefik.http.routers"
    s = "traefik.http.services"
    m = "traefik.http.middlewares"
    for domain_name, domain in http_config["domains"].items():
        router_name = f"{service_name}-{slugify(domain_name)}"
        if domain_name.startswith("*."):
            # for wildcard, set minimal priority and ask for *.name certificate
            top_domain = domain_name[2:]
            labels[f"{r}.{router_name}.priority"] = "1"
            labels[
                f"{r}.{router_name}.rule"
            ] = f"HostRegexp(`{top_domain}`, `{{subdomain:.*}}.{top_domain}`)"
            if ssl:
                labels[f"{r}.{router_name}.tls.domains[0].main"] = top_domain
                labels[f"{r}.{router_name}.tls.domains[0].sans"] = domain_name
        else:
            labels[f"{r}.{router_name}.rule"] = f"Host(`{domain_name}`)"

        if ssl:
            labels[f"{r}.{router_name}.entrypoints"] = "https"
            labels[f"{r}.{router_name}.tls.certresolver"] = "letsencrypt"
        else:
            labels[f"{r}.{router_name}.entrypoints"] = "http"

        labels[f"{r}.{router_name}.service"] = router_name
        labels[f"{s}.{router_name}.loadbalancer.server.port"] = str(port)

        # middleware
        middlewares = []
        md_name = f"chain-{service_name}-{slugify(domain_name)}"
        

        # IP allow
        md_ial_name = f"al-{service_name}-{slugify(domain_name)}"
        allow_list = ",".join(str(ip) for ip in domain["ip-allow-list"])
        if allow_list:
            labels[f"{m}.{md_ial_name}.ipwhitelist.sourcerange"] = allow_list
            middlewares.append(md_ial_name)

        # BasicAuth
        md_ba_name = f"ba-{service_name}-{slugify(domain_name)}"
        auth_list = ",".join(str(ip) for ip in domain["basic-auth"])
        if auth_list:
            labels[f"{m}.{md_ba_name}.basicauth.users"] = auth_list
            middlewares.append(md_ba_name)

        if middlewares:
            labels[f"{m}.{md_name}.chain.middlewares"] = ",".join(middlewares)
            labels[f"{r}.{router_name}.middlewares"] = f"{md_name}@docker"
    return labels
